%%  PROGRAM DESCRIPTION
%
% This program implements a matrix-based, regularized,
% non-iterative solution of transport of intensity (TIE) to compute the
% pseudo-phase of a sample from a sequence of 3 intensity images.
% The program performs steps 1-4 in about 220 milliseconds on a 1 megapixel
% image on a desktop computer with a i7-4770 3.4GHz 4-core processor.
% There is an option to accelerate the calculation with a GPU. Using a
% GeForce GTX-670 processor with 1344 cores, the time for the 1 megapixel
% image is reduced to 20 milliseconds. Checking the answer adds a few
% milliseconds (step 5). Plotting and saving data to disk (step 6) adds 
% a few hundred milliseconds.
%
% If you use this code in your research, please cite both the original
% paper in which the method was presented and the code's figshare page:
%
% [1] J. Zhu, R. Zhou, L. Zhang, B. Ge, C. Luo, and L. L. Goddard, 
% "Regularized pseudo-phase imaging for inspecting and sensing
% nanoscale features," to be published (2018).
%
% [2] J. Zhu and L.L. Goddard, figshare (2017) [retrieved {enter date}],
% https://figshare.com/articles/Robust_TIE_Solver_m/5504086.
%
%% HOW TO USE THIS PROGRAM
%
% Each image must be a 2d array of doubles. The images must have 1-1 aspect
% ratio, i.e. a pixel shift in x corresponds to the same physical distance
% as a pixel shift in y. If for some reason your images are not 1-1 aspect
% ratio, you must either stretch the images or resample them before using
% this program.
%
% The matrix operations require square matrices. Thus, rectangular input
% images will automatically be cropped into a square image of the central
% region.
%
% This program uses 2 functions. The first function does initialization.
% The second function can be put inside a loop to process multiple images.
%
% [Q, Qinv, Z, init_vars] = Initialize_TIE_Solver(use_gpu, ...
%    use_single, rows, columns, reg_factor, Neumann);
% [phase, I0_crop, difference] = Perform_TIE_Solver(I0_pos, I0, I0_neg, ...
%    um_per_pixel, dz, k0, Q, Qinv, Z, init_vars, out_file);
%
%% INPUTS
%
% use_gpu - set to true if you want to use the GPU accelerated version
% use_single - set true to use single precision (faster but less accurate)
% check_answer - set true to see the error in the TIE equation
% plot_data - set true to plot the results in figures
% save_data - set true to save data to out_file
% rows - the number of rows in the image
% columns - the number of columns in the image
% reg_factor1 - a small parameter (usually 5e-4)
% reg_factor2 - a small parameter (usually 5e-5)
%   These regularization factors need to be carefully chosen. If they are
%   too small, then measurement errors remain amplified. If they are too 
%   large, features get filtered away. The optimal choice for reg_factor1
%   depends both on the amount of noise and on the size of the image. The
%   choice for reg_factor2 can be half the size of the quantization error
%   in the measurement, e.g., for 12-bit acquisition, this error is 1e-4
%   which gives reg_factor2 = 5e-5.
% Neumann - set boundary conditions: Neumann (true) or Dirichlet (false)
%   If Neumann is set as true (preferred), the program updates the
%   tridiagonal matrix T under the assumption that the images are 
%   continuous right outside the boundaries. If Neumann is set to false, 
%   the images are assumed to be zero outside the boundary (Dirichlet). 
%   This may lead to strong edge effects if this condition is not
%   specifically engineered in the experiment.
%
% I0_pos - intensity image at a distance dz above the image plane
% I0 - intensity image at the image plane
% I0_neg - intensity image at a distance dz below the image plane
% um_per_pixel - the size (in microns) of a pixel in the sample plane
% dz - the distance in z (in microns) between each of the three images
%   Note, if you translate the camera instead of translating the sample,
%   you would need to enter the physical size of a pixel on the camera
%   plane for um_per_pixel and enter the camera translation distance as dz.
% k0 - the wave vector, given by 2*pi*ref_index/lambda where ref_index is
%   the refractive index of the ambient environment and lambda is the
%   central wavelength (in microns) of the white light source
% crop_rows - intermediate variable from initialization
% crop_cols - intermediate variable from initialization
% out_file - name of the output file to write the results
%
%% OUTPUTS
%
% phase - computed output pseudo-phase image in units of radians
% I0_crop - cropped intensity image at the best focus plane
% TIE_RHS - the right hand side of the TIE (namely the derivative in z)
% F - the intermediate matrix whose value is div(I(x,y)*grad(phase(x,y)))
%
%% COMMENTS FOR THE PROGRAM'S MAIN STEPS TO CALCULATE PSEUDO-PHASE
%
% ## Step 0: One-time initialization of matrices (depends on crop window)
%
% ## Step 1: Crop images to make them square, and if necessary single
% precision
%
% ## Step 2: Solve the TIE to find intermediate matrix F ##
% The 3 separate equations in the paper are combined here into 1 step in
% consideration of speed and memory. The 3 steps would have been:
% H = Qinv * TIE_RHS * Q; K = H .* Z; F = Q * K * Qinv;
%
% ## Step 3: Compute divergence of (grad F)/I0 ##
%
% ## Step 4: Solve Poisson to find pseudo-phase ##
% With these two second derivatives computed, the Laplacian of phi can be
% written as: Del2_phi = grad2_phi_x + grad2_phi_y;
% This intermediate result and the corresponding 3 intermediate steps
% are again combined for speed and memory consierations:
% H = Qinv * Del2_phi * Q; K = H .* Z; phase = Q * K * Qinv;
%
% ## Step 5: Calculate the difference RHS minus LHS of TIE ##
%
% ## Step 6: Plot and save results ##
%
%% Acknowledgment:
% Images of red blood cells (courtesy of Baoliang Ge, MIT) are used as an
% example for running the program.
%
% Creative Commons License:
% Attribution-NonCommercial-NoDerivatives 4.0 International
% Copyright by Jinlong Zhu and Lynford L. Goddard
% Photonic Systems Lab, University of Illinois at Urbana-Champaign.
% Version 9, 10-08-2018.
%
% ####################################################################

%% ########### Example for the main code to use this program ###########
clc;close all

%% ## Specify inputs ##
use_gpu = false;
use_single = false;
check_answer = true;
plot_data = true;
save_data = true;
out_file = 'TIE_results.mat';

%% ## Specify experimental conditions ##
um_per_pixel = 1/18;  % size of an pixel in sample plane (in microns)
dz = 1;               % translation distance of the sample (in microns)
Neumann = true;       % Neumann (true) or Dirichlet (false) boundaries
lambda = 0.575;       % vacuum wavelength (in microns)
ref_index = 1;        % index of ambient environment
k0 = 2*pi*ref_index/lambda;  % wave vector
reg_factor1 = 5e-4;   % regularization factor1 (1e-6 for RBC, 5e-4 for LER)
reg_factor2 = 5e-5;   % regularization factor2 (5e-5 for 12-bit images)

%% ## Specify data to be processed ##
I0_pos = im2double(rgb2gray(imread('LER-Z1.png')));
I0     = im2double(rgb2gray(imread('LER-Z0.png')));
I0_neg = im2double(rgb2gray(imread('LER-Z-1.png')));
rows = length(I0(:,1));
columns = length(I0(1,:));

%% Roughly estimate the experimental error by comparing 2 estimates for I0
noise = std2(I0-0.5*(I0_pos+I0_neg))/max(max(I0))

%% Initialize solver and call functions
[Q, Qinv, Z, init_vars] = Initialize_TIE_Solver(use_gpu, ...
    use_single, check_answer, plot_data, save_data, rows, columns, ...
    reg_factor1, reg_factor2, Neumann);
[phase, I0_crop, difference] = Perform_TIE_Solver(I0_pos, I0, I0_neg, ...
    um_per_pixel, dz, k0, Q, Qinv, Z, init_vars, out_file);