function [phase, I0_crop, difference] = Perform_TIE_Solver(I0_pos, I0, ...
    I0_neg, um_per_pixel, dz, k0, Q, Qinv, Z, init_vars, out_file)
%% Step 1a: Recover variables
use_gpu = init_vars.use_gpu;
use_single = init_vars.use_single;
crop_cols = init_vars.crop_cols;
crop_rows = init_vars.crop_rows;
reg_factor1 = init_vars.reg_factor1;
reg_factor2 = init_vars.reg_factor2;
check_answer = init_vars.check_answer;
plot_data = init_vars.plot_data;
save_data = init_vars.save_data;

%% Step 1b: Convert images to single precision if necessary
if use_single
    I0_pos = single(I0_pos);
    I0 = single(I0);
    I0_neg = single(I0_neg);
end

%% Step 1c: Crop input images to make them square
dx2 = um_per_pixel^2;    % dx^2 = dy^2 in microns (used for 2D Laplacian)
rhs_const = -k0 * dx2 / (2*dz); % scaling constant for RHS of TIE
if use_gpu
    I0_crop = gpuArray(I0(crop_rows,crop_cols));
    TIE_RHS = gpuArray(rhs_const * (I0_pos(crop_rows,crop_cols) - ...
        I0_neg(crop_rows,crop_cols)));
else
    I0_crop = I0(crop_rows,crop_cols);
    TIE_RHS = rhs_const * (I0_pos(crop_rows,crop_cols) - ...
        I0_neg(crop_rows,crop_cols));
end

%% Steps 2-4: Recover the phase. Step 1c above plus these 5 lines can be
% put inside a for loop to execute on multiple measurements taken with the
% same configuration.
F = Q * ((Qinv * TIE_RHS * Q) .* Z) * Qinv;                       % Step 2
[grad1x_F,grad1y_F] = gradient(F);                                % Step 3a
[grad2x_phi,~] = gradient(grad1x_F./(I0_crop + reg_factor2));     % Step 3b
[~,grad2y_phi] = gradient(grad1y_F./(I0_crop + reg_factor2));     % Step 3c
phase = Q * ((Qinv * (grad2x_phi + grad2y_phi) * Q) .* Z) * Qinv; % Step 4

%% Step 5: Check the agreement of the TIE solution
if check_answer
    [grad1_phase_x,grad1_phase_y] = gradient(phase);
    [grad2_F_x,~] = gradient(I0_crop.*grad1_phase_x);
    [~,grad2_F_y] = gradient(I0_crop.*grad1_phase_y);
    difference = TIE_RHS - (grad2_F_x + grad2_F_y);
else
    difference = 0*TIE_RHS;
end

%% Step 6a: Draw the images
if plot_data
    s1='Raw Intensity Image at z0';
    s2=sprintf('Robust Pseudo-Phase, epsilon=%.1e',reg_factor1);
    s3=sprintf('TIE error, rms=%.1e, max=%.1e',std2(difference)/max(max(abs(TIE_RHS))),...
        max(max(abs(difference)))/max(max(abs(TIE_RHS))));
    figure, imagesc(I0_crop), axis image, colormap jet; colorbar, title(s1)
    figure, imagesc(phase), axis image, colormap jet; colorbar, title(s2)
    figure, imagesc(difference), axis image, colormap jet; colorbar, title(s3)
end

%% Step 6b: Save the output data
if save_data
    save(out_file,'I0_crop','phase');
end
end